unit Rectangles;
  { provides ADT for axis-parallel grid rectangles }

interface

type
  TRectangle = class(TObject)  // axis-parallel grid rectangles
  private (* ignore implementation details *)
    // fields
    FXL: Integer; // lower X coordinate
    FYL: Integer; // lower Y coordinate
    FXH: Integer; // higher X coordinate
    FYH: Integer; // higher Y coordinate

  public
    // construction
    constructor Create(AXL, AYL, AXH, AYH: Integer);
      // pre: AXL <= AXH  /\  AYL <= AYH
      // post: XL=AXL /\ YL=AYL /\ XH=AXH /\ YH=AYH

    // basic queries
    function XL: Integer; // lower X coordinate 
    function YL: Integer; // lower Y coordinate
    function XH: Integer; // higher X coordinate
    function YH: Integer; // higher Y coordinate

    // invariants
    // WellFormed: XL <= XH  /\  YL <= YH

    // derived queries
    function Contains(AX, AY: Integer): Boolean;
      // pre: true
      // ret: XL <= AX <= XH  /\  YL <= AY <= YH
    function Intersects(const ARectangle: TRectangle): Boolean;
      // pre: true
      // ret: Self intersects ARectangle

    //commands
    procedure SetXL(AX: Integer);
      // pre: AX <= XH
      // effect: XL := AX
    procedure SetYL(AY: Integer);
      // pre: AY <= YH
      // effect: YL := AY
    procedure SetXH(AX: Integer);
      // pre: XL <= AX
      // effect: XH := AX
    procedure SetYH(AY: Integer);
      // pre: YL <= AY
      // effect: YH := AY
    procedure Shift(AX, AY: Integer);
      // pre: true
      // effect: XL, YL, XH, YL := XL+AX, YL+AY, XH+AX, YH+AY
    procedure Intersect(const ARectangle: TRectangle);
      // pre: Intersects(ARectangle)
      // effect: Self := Self intersection ARectangle
    procedure Hull(const ARectangle: TRectangle);
      // pre: true
      // effect: Self := smallest rectangle enclosing
      //                 Self and ARectangle
  end; { class TRectangle }

implementation

uses
  Math; { for Min and Max }

constructor TRectangle.Create(AXL, AYL, AXH, AYH: Integer);
  begin
    (* pre not checked *)
    inherited Create
  ; FXL := AXL
  ; FYL := AYL
  ; FXH := AXH
  ; FYH := AYH
  end; { Create }

function TRectangle.XL: Integer;
  begin
    Result := FXL
  end; { XL }

function TRectangle.YL: Integer;
  begin
    Result := FYL
  end; { YL }

function TRectangle.XH: Integer;
  begin
    Result := FXH
  end; { XH }

function TRectangle.YH: Integer;
  begin
    Result := FYH
  end; { YH }

function TRectangle.Contains(AX, AY: Integer): Boolean;
  begin
    Result := (XL <= AX) and (AX <= XH) and
              (YL <= AY) and (AY <= YH)
  end; { Contains }

function TRectangle.Intersects(const ARectangle: TRectangle): Boolean;
  begin
    Result := (XL <= ARectangle.XH)  // not Self right of ARectangle
          and (ARectangle.XL <= XH)  // not Self left of ARectangle
          and (YL <= ARectangle.YH)  // not Self above ARectangle
          and (ARectangle.YL <= YH)  // not Self below ARectangle
  end; { Intersects }

procedure TRectangle.SetXL(AX: Integer);
  begin
    (* pre not checked *)
    FXL := AX
  end; { SetXL }

procedure TRectangle.SetYL(AY: Integer);
  begin
    (* pre not checked *)
    FYL := AY
  end; { SetYL }

procedure TRectangle.SetXH(AX: Integer);
  begin
    (* pre not checked *)
    FXH := AX
  end; { SetXH }

procedure TRectangle.SetYH(AY: Integer);
  begin
    (* pre not checked *)
    FYH := AY
  end; { SetYH }

procedure TRectangle.Shift(AX, AY: Integer);
  begin
    FXL := FXL + AX
  ; FYL := FYL + AY
  ; FXH := FXH + AX
  ; FYH := FYH + AY
  end; { Shift }

procedure TRectangle.Intersect(const ARectangle: TRectangle);
  begin
    (* pre not checked *)
    SetXL(Max(XL, ARectangle.XL))
  ; SetYL(Max(YL, ARectangle.YL))
  ; SetXH(Min(XH, ARectangle.XH))
  ; SetYH(Min(YH, ARectangle.YH))
  end; { Intersect }

procedure TRectangle.Hull(const ARectangle: TRectangle);
  begin
    SetXL(Min(XL, ARectangle.XL))
  ; SetYL(Min(YL, ARectangle.YL))
  ; SetXH(Max(XH, ARectangle.XH))
  ; SetYH(Max(YH, ARectangle.YH))
  end; { Hull }

end.
